"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.dockerContainers = exports.runWatcher = exports.resetDb = exports.checkOpenrestyInitiated = exports.checkMigrationsInitiated = exports.checkIsAppDir = exports.notEmptyString = exports.dirExists = exports.fileExists = exports.sqitchDeploy = exports.checkPostgresConnection = exports.runCmd = void 0;

var _child_process = _interopRequireDefault(require("child_process"));

var _fs = _interopRequireDefault(require("fs"));

var _os = _interopRequireDefault(require("os"));

var _chokidar = _interopRequireDefault(require("chokidar"));

var _string_decoder = require("string_decoder");

var _env = require("./env.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const runCmd = (cmd, params, options = {}, silent = false, exit_on_error = false, use_async = false) => {
  if (_env.USE_DOCKER_IMAGE && [_env.PSQL_CMD, _env.SQITCH_CMD, _env.PG_DUMP_CMD, _env.PG_DUMPALL_CMD, _env.JAVA_CMD].indexOf(cmd) !== -1) {
    //alter the command to run in docker
    let w = options && options.cwd ? options.cwd.replace(_env.APP_DIR, _env.DOCKER_APP_DIR) : _env.DOCKER_APP_DIR;
    let e = options && options.env ? Object.keys(options.env).reduce(function (acc, key) {
      acc.push('-e');
      acc.push(key + '=' + options.env[key].replace(_env.APP_DIR, _env.DOCKER_APP_DIR));
      return acc;
    }, []) : [];
    let u = [],
        hostAppDir = _env.APP_DIR;

    if (_env.PLATFORM == 'linux') {
      let {
        uid,
        gid
      } = _os.default.userInfo();

      u = ['-u', `${uid}:${gid}`, '--env', "USERNAME=root"];
    }

    if (_env.PLATFORM == 'win32' && _env.DOCKER_HOST_OS != "Docker for Windows") {
      // docker is running in virtualbox VM so we try to adjust the path to match the shared path
      hostAppDir = '/' + hostAppDir.replace(/([A-Z]):/, function (s) {
        return s.toLocaleLowerCase().replace(':', '');
      }).replace(/\\/g, '/');
    }

    let ef = fileExists(`.env`) ? ['--env-file', `${_env.APP_DIR}/.env`] : [];
    let p = ['run', '--net', 'host', '--rm', '-w', w, '-v', `${hostAppDir}:${_env.DOCKER_APP_DIR}`].concat(e).concat(u).concat(ef).concat([_env.DOCKER_IMAGE, cmd]).concat(params.map(v => v.replace(_env.APP_DIR, _env.DOCKER_APP_DIR)));
    cmd = 'docker';
    params = p;
  }

  if (!(silent || use_async)) {
    options.stdio = ['ignore', 1, 2];
  }

  let spawn = use_async ? _child_process.default.spawn : _child_process.default.spawnSync;
  let pr = spawn(cmd, params, options);

  if (!use_async && exit_on_error && pr.status != 0) {
    process.exit(pr.status);
  }

  return pr;
};

exports.runCmd = runCmd;

const checkPostgresConnection = (connection_string, pass, exit_on_error = true) => {
  console.log(`Checking PostgreSQL connection info (${connection_string})`);
  var env = Object.create(process.env);
  env.PGPASSWORD = pass;
  const random_number = Math.floor(Math.random() * 100).toString();
  const params = ['--quiet', '--tuples-only', '-c', `SELECT ${random_number}`, connection_string];
  let result = runCmd(_env.PSQL_CMD, params, {
    env: env
  }, true, false).stdout.toString().replace(/\s/g, '');

  if (random_number !== result) {
    console.log(" ");
    console.log(`Could not verify the connection to PostgreSQL`);

    if (exit_on_error) {
      process.exit(1);
    } else {
      return false;
    }
  }

  return true;
};

exports.checkPostgresConnection = checkPostgresConnection;

const sqitchDeploy = url => runCmd(_env.SQITCH_CMD, ["deploy", url], {
  cwd: _env.MIGRATIONS_DIR
}, false, true);

exports.sqitchDeploy = sqitchDeploy;

const fileExists = path => _fs.default.existsSync(path) && _fs.default.statSync(path).isFile();

exports.fileExists = fileExists;

const dirExists = path => _fs.default.existsSync(path) && _fs.default.statSync(path).isDirectory(); // options.key from commander returns bool if a value is not specified(e.g. subzero cloud login -u, options.username gives true), so make sure is a string


exports.dirExists = dirExists;

const notEmptyString = s => typeof s == 'string' && s.trim().length;

exports.notEmptyString = notEmptyString;

const checkIsAppDir = () => {
  if (!fileExists(`.env`)) {
    console.log("Error: ".red + ".env file does not exist");
    console.log("Please run this command in a directory that contains a subzero/postgrest project.\nYou can create a poject with " + "`subzero base-project`".white);
    process.exit(0);
  }
};

exports.checkIsAppDir = checkIsAppDir;

const checkMigrationsInitiated = () => {
  if (!(dirExists(_env.MIGRATIONS_DIR) && fileExists(`${_env.MIGRATIONS_DIR}/sqitch.plan`))) {
    console.log("Error: ".red + "database migrations not initiated");
    console.log("Please run `subzero migrations init` before trying to deploy the code");
    process.exit(0);
  }
};

exports.checkMigrationsInitiated = checkMigrationsInitiated;

const checkOpenrestyInitiated = () => {
  if (!(dirExists(_env.OPENRESTY_DIR) && fileExists(`${_env.OPENRESTY_DIR}/Dockerfile`))) {
    console.log("Error: ".red + "Dockerfile for custom OpenResty image missing");
    console.log(`Please create ${_env.OPENRESTY_DIR}/Dockerfile`);
    process.exit(0);
  }
};

exports.checkOpenrestyInitiated = checkOpenrestyInitiated;
const decoder = new _string_decoder.StringDecoder('utf8');

const resetDb = (containers, logger) => {
  const psql = runSql(['postgres', '-c', `SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname = '${_env.DB_NAME}';`, '-c', `DROP DATABASE if exists ${_env.DB_NAME};`, '-c', `CREATE DATABASE ${_env.DB_NAME};`], logger);
  psql.stderr.on('data', data => logger.log(decoder.write(data)));
  psql.on('close', code => {
    const _psql = runSql([_env.DB_NAME, '-f', _env.DB_DIR + 'init.sql'], logger);

    _psql.stderr.on('data', data => logger.log(decoder.write(data)));

    _psql.on('close', code => {
      if (code == 0) {
        if (containers['postgrest']) {
          sendHUP(containers['postgrest'].name);
        }

        if (containers['openresty']) {
          sendHUP(containers['openresty'].name);
        }
      }
    });
  });
  return psql;
};

exports.resetDb = resetDb;

const sendHUP = containerName => _child_process.default.spawn('docker', ['kill', '-s', 'HUP', containerName]);

const runSql = commands => {
  const connectParams = ['-U', _env.SUPER_USER, '-h', 'localhost', '--set', 'DIR=' + _env.DB_DIR, '--set', 'ON_ERROR_STOP=1'];
  var env = Object.create(process.env);
  env.PGPASSWORD = _env.SUPER_USER_PASSWORD;
  env.DIR = _env.DB_DIR;
  return runCmd(_env.PSQL_CMD, connectParams.concat(commands), {
    env: env
  }, undefined, undefined, true);
};

const runWatcher = (containers, logger, watcherReadyCb, reloadStartCb, reloadEndCb) => {
  return _chokidar.default.watch(_env.WATCH_PATTERNS, {
    ignored: _env.APP_DIR + '/**/tests/**'
  }).on('change', path => {
    const relPath = path.replace(_env.APP_DIR, '.');
    reloadStartCb(relPath);

    if (path.endsWith('.sql')) {
      resetDb(containers, logger).on('close', reloadEndCb);
    } else {
      if (containers['openresty']) sendHUP(containers['openresty'].name).on('close', reloadEndCb);else reloadEndCb();
    }
  }).on('ready', watcherReadyCb);
};

exports.runWatcher = runWatcher;

const dockerContainers = () => {
  const containers = _child_process.default.execSync(`docker ps -a -f name=${_env.COMPOSE_PROJECT_NAME} --format "{{.Names}}"`).toString('utf8').trim().split("\n");

  return containers.reduce((acc, containerName) => {
    let key = containerName.replace(_env.COMPOSE_PROJECT_NAME, '').replace('1', '').replace(/_/g, '');
    acc[key] = {
      name: containerName,
      title: containerName.replace(_env.COMPOSE_PROJECT_NAME + '_', '').replace(/_1$/, '')
    };
    return acc;
  }, {});
};

exports.dockerContainers = dockerContainers;