#!/usr/bin/env node
"use strict";

var _commander = _interopRequireDefault(require("commander"));

var _child_process = _interopRequireDefault(require("child_process"));

var _fs = _interopRequireDefault(require("fs"));

var _sleep = _interopRequireDefault(require("sleep"));

var _common = require("./common.js");

var _net = require("net");

var _env = require("./env.js");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const TMP_DIR = `${_env.MIGRATIONS_DIR}/tmp`;
const INITIAL_FILE_NAME = "initial";
const SQITCH_CONF = `${_env.MIGRATIONS_DIR}/sqitch.conf`;
const INIT_SQL_FILENAME = '0000000000-setup.sql';

const getFreePort = async () => {
  return new Promise((resolve, reject) => {
    const server = (0, _net.createServer)();
    server.on('error', reject);
    server.on('listening', () => (port => server.close(() => resolve(port)))(server.address().port));
    server.listen(0);
  });
};

const initMigrations = async (debug, dbDockerImage) => {
  if (_fs.default.existsSync(_env.MIGRATIONS_DIR)) {
    console.log(`Migrations directory already exists: ${_env.MIGRATIONS_DIR}`);
    process.exit(-1);
  }

  const name = INITIAL_FILE_NAME;
  const migrationNumber = getMigrationNumber();

  _fs.default.mkdirSync(_env.MIGRATIONS_DIR);

  _fs.default.mkdirSync(TMP_DIR);

  const {
    containerName: devDbcontainerName,
    dbUri: devDbUri,
    initSqlFile: devInitSqlFile
  } = await getTempPostgres(_env.DB_DIR, debug, dbDockerImage);
  const devDbContianerReady = waitForDbContainer(devDbcontainerName, devInitSqlFile, debug);

  if (!devDbContianerReady) {
    stopContainer(devDbcontainerName);
    process.exit(-1);
  }

  dumpSchema(devDbUri, `${TMP_DIR}/${migrationNumber}-dev-${name}.sql`, true);
  stopContainer(devDbcontainerName);

  _fs.default.closeSync(_fs.default.openSync(`${TMP_DIR}/${migrationNumber}-prod-${name}.sql`, 'w'));

  initSqitch();
  addSqitchMigration(`${migrationNumber}-${name}`);
  apgdiffToFile(`${TMP_DIR}/${migrationNumber}-dev-${name}.sql`, `${TMP_DIR}/${migrationNumber}-prod-${name}.sql`, `${TMP_DIR}/${migrationNumber}-revert-${name}.sql`);
  console.log(`Copying ${TMP_DIR}/${migrationNumber}-dev-${name}.sql to ${_env.MIGRATIONS_DIR}/deploy/${migrationNumber}-${name}.sql`);

  _fs.default.copyFileSync(`${TMP_DIR}/${migrationNumber}-dev-${name}.sql`, `${_env.MIGRATIONS_DIR}/deploy/${migrationNumber}-${name}.sql`);

  _fs.default.unlinkSync(`${TMP_DIR}/${migrationNumber}-dev-${name}.sql`);

  _fs.default.unlinkSync(`${TMP_DIR}/${migrationNumber}-prod-${name}.sql`);

  _fs.default.unlinkSync(`${TMP_DIR}/${migrationNumber}-revert-${name}.sql`);
};

const addMigration = async (name, note, diff, dryRun, debug, dbUri, dbDockerImage, diffRoles) => {
  if (!_fs.default.existsSync(SQITCH_CONF) || !_fs.default.statSync(SQITCH_CONF).isFile()) {
    console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", SQITCH_CONF);
    process.exit(-1);
  }

  if (dbUri) {
    (0, _common.checkPostgresConnection)(dbUri, '');
  }

  const migrationNumber = getMigrationNumber(),
        tmpDevSql = `${TMP_DIR}/${migrationNumber}-dev-${name}.sql`,
        tmpProdSql = `${TMP_DIR}/${migrationNumber}-prod-${name}.sql`,
        tmpDeploySql = `${TMP_DIR}/${migrationNumber}-deploy-${name}.sql`,
        tmpRevertSql = `${TMP_DIR}/${migrationNumber}-revert-${name}.sql`;

  if (diff) {
    if (!_fs.default.existsSync(TMP_DIR)) _fs.default.mkdirSync(TMP_DIR);
    const {
      containerName: devDbcontainerName,
      dbUri: devDbUri,
      initSqlFile: devInitSqlFile
    } = await getTempPostgres(_env.DB_DIR, debug, dbDockerImage);
    const devDbContianerReady = waitForDbContainer(devDbcontainerName, devInitSqlFile, debug);

    if (!devDbContianerReady) {
      stopContainer(devDbcontainerName);
      process.exit(-1);
    }

    dumpSchema(devDbUri, tmpDevSql, diffRoles);
    stopContainer(devDbcontainerName);

    if (dbUri) {
      dumpSchema(dbUri, tmpProdSql, diffRoles);
    } else {
      const {
        containerName: prodDbcontainerName,
        dbUri: prodDbUri,
        initSqlFile: prodInitSqlFile
      } = await getTempPostgres(`${_env.MIGRATIONS_DIR}/deploy`, debug, dbDockerImage);
      const prodDbContianerReady = waitForDbContainer(prodDbcontainerName, prodInitSqlFile, debug);

      if (!prodDbContianerReady) {
        stopContainer(prodDbcontainerName);
        process.exit(-1);
      }

      dumpSchema(prodDbUri, tmpProdSql, diffRoles);
      stopContainer(prodDbcontainerName);
    }

    apgdiffToFile(tmpDevSql, tmpProdSql, tmpRevertSql);
    apgdiffToFile(tmpProdSql, tmpDevSql, tmpDeploySql);
  }

  if (!dryRun) {
    addSqitchMigration(`${migrationNumber}-${name}`, note);

    if (diff) {
      if (_fs.default.existsSync(tmpDeploySql)) {
        _fs.default.copyFileSync(tmpDeploySql, `${_env.MIGRATIONS_DIR}/deploy/${migrationNumber}-${name}.sql`);
      } else {
        console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", `${migrationNumber}-deploy-${name}.sql`);
      }

      if (_fs.default.existsSync(tmpRevertSql)) {
        _fs.default.copyFileSync(tmpRevertSql, `${_env.MIGRATIONS_DIR}/revert/${migrationNumber}-${name}.sql`);
      } else {
        console.log("\x1b[31mError:\x1b[0m the file '%s' does not exist", `${migrationNumber}-revert-${name}.sql`);
      }

      console.log(`\x1b[31mATTENTION:\x1b[0m Make sure you check deploy/${migrationNumber}-${name}.sql for correctness, statement order is not handled!`);
    } else {
      console.log('Creating empty migration');
    }
  } else {
    console.log("\nCurrent migration DDL");
    console.log("=====================================================");

    if (_fs.default.existsSync(tmpDeploySql)) {
      console.log(_fs.default.readFileSync(tmpDeploySql, 'utf-8'));
    }
  }

  if (!debug) {
    if (_fs.default.existsSync(tmpDevSql)) {
      _fs.default.unlinkSync(tmpDevSql);
    }

    if (_fs.default.existsSync(tmpProdSql)) {
      _fs.default.unlinkSync(tmpProdSql);
    }

    if (_fs.default.existsSync(tmpRevertSql)) {
      _fs.default.unlinkSync(tmpRevertSql);
    }

    if (_fs.default.existsSync(tmpDeploySql)) {
      _fs.default.unlinkSync(tmpDeploySql);
    }
  }
};

const getMigrationNumber = () => new Date().toISOString().slice(0, 19).replace(/[^0-9]/g, '');

const stopContainer = name => {
  (0, _common.runCmd)("docker", ["stop", name], undefined, true);
  (0, _common.runCmd)("docker", ["rm", name], undefined, true);
};

const writeInitSql = file => {
  _fs.default.writeFileSync(file, ['-- custom setup sql', `create role "${_env.DB_USER}" with login password '${_env.DB_PASS}';`].join("\n"));
};

const getTempPostgres = async (sqlDir, debug, dbDockerImage) => {
  const name = 'temp_postgres_' + Math.random().toString(36).substr(2, 5);
  const port = await getFreePort();
  const initSqlFile = `${sqlDir}/${INIT_SQL_FILENAME}`;
  writeInitSql(initSqlFile);
  let pgVersion = getPgVersion(_env.DEV_DB_URI);
  let pgDockerImage = dbDockerImage ? dbDockerImage : `postgres:${pgVersion}`;
  console.log(`Starting temporary PostgreSQL database with the version matching your dev database (${pgDockerImage})`);
  console.log(`on port ${port} with SQL from ${sqlDir}`);
  const cmd = 'docker',
        params = ["run", "-d", "--name", name, "-p", port + ":5432", //# env vars specific to postgres image used on first boot
  "-e", `POSTGRES_USER=${_env.SUPER_USER}`, "-e", `POSTGRES_PASSWORD=${_env.SUPER_USER_PASSWORD}`, "-e", `POSTGRES_DB=${_env.DB_NAME}`, //# env vars useful for our sql scripts
  "-e", `SUPER_USER=${_env.SUPER_USER}`, "-e", `SUPER_USER_PASSWORD=${_env.SUPER_USER_PASSWORD}`, "-e", `DB_NAME=${_env.DB_NAME}`, "-e", `DB_USER=${_env.DB_USER}`, "-e", `DB_PASS=${_env.DB_PASS}`, "-e", `DB_ANON_ROLE=${process.env.DB_ANON_ROLE}`, "-e", `DEVELOPMENT=${process.env.DEVELOPMENT}`, "-e", `JWT_SECRET=${process.env.JWT_SECRET}`, "-v", `${sqlDir}:/docker-entrypoint-initdb.d`, pgDockerImage];

  if (debug) {
    console.log("Starting temp contianer with:");
    console.log(`${cmd} ${params.join(' ')}`);
  }

  (0, _common.runCmd)(cmd, params);
  return {
    containerName: name,
    dbUri: `postgres://${_env.SUPER_USER}:${_env.SUPER_USER_PASSWORD}@${_env.LOCALHOST}:${port}/${_env.DB_NAME}`,
    initSqlFile: initSqlFile
  };
};

const waitForDbContainer = (name, initSqlFile, debug) => {
  console.log(`Waiting for ${name} container to load`);
  let finishedLoading = false;
  let timestamp = 0;
  let iterations = 0;
  const maxIterations = 60;

  while (!finishedLoading) {
    iterations = iterations + 1;

    if (iterations > maxIterations) {
      break;
    }

    let p = _child_process.default.spawnSync('docker', ['logs', '--since', timestamp, name]);

    timestamp = Math.floor(new Date() / 1000);
    p.output.forEach(v => {
      if (v) {
        if (v.toString().indexOf('PostgreSQL init process complete; ready for start up.') !== -1) {
          finishedLoading = true;
        }
      }
    });

    _sleep.default.sleep(1);
  }

  _sleep.default.sleep(1);

  if (_fs.default.existsSync(initSqlFile)) {
    _fs.default.unlinkSync(initSqlFile);
  }

  ;

  if (finishedLoading) {
    console.log('PostgreSQL init process complete; ready for start up.');

    if (debug) {
      (0, _common.runCmd)('docker', ['logs', name]);
    }
  } else {
    (0, _common.runCmd)('docker', ['logs', '--tail', '30', name]);
    console.log('Gave up on waiting for db. The last 30 lines of log are above.');

    if (_fs.default.existsSync(initSqlFile)) {
      _fs.default.unlinkSync(initSqlFile);
    }

    ;
  }

  return finishedLoading;
};

const initSqitch = () => (0, _common.runCmd)(_env.SQITCH_CMD, ["init", _env.DB_NAME, "--engine", "pg"], {
  cwd: _env.MIGRATIONS_DIR
});

const addSqitchMigration = (name, note) => (0, _common.runCmd)(_env.SQITCH_CMD, ["add", name, "-n", note || `Add ${name} migration`], {
  cwd: _env.MIGRATIONS_DIR
});

const dumpSchema = (dbUri, file, diffRoles) => {
  console.log(`Writing database dump to ${file}`);
  const replace_superuser = new RegExp(`GRANT ([a-z0-9_-]+) TO ${_env.SUPER_USER}`, "gi");

  if (diffRoles) {
    (0, _common.runCmd)(_env.PG_DUMPALL_CMD, ['-f', `${file}.roles`, '--roles-only', '-d', dbUri]);
  }

  (0, _common.runCmd)(_env.PG_DUMP_CMD, [dbUri, '-f', `${file}.schema`, '--schema-only']);
  let data = [!diffRoles ? '' : _fs.default.readFileSync(`${file}.roles`, 'utf-8').split("\n").filter(ln => _env.IGNORE_ROLES.map(r => ln.indexOf('ROLE ' + r)).every(p => p == -1)) //filter out line referring to ignored roles
  .map(ln => ln.replace(` GRANTED BY ${_env.SUPER_USER}`, '')) //remove unwanted string
  .filter(ln => ln.indexOf('ALTER ROLE') == -1) //RDS does not allow this
  .map(ln => ln.replace(replace_superuser, 'GRANT $1 TO current_user')).join("\n"), _fs.default.readFileSync(`${file}.schema`, 'utf-8').split("\n").filter(ln => ln.indexOf('COMMENT ON EXTENSION') == -1) //RDS doew not allow this
  .filter(ln => ln.indexOf(`OWNER TO ${_env.SUPER_USER};`) == -1) //don't keep owner info when the owner is privileges
  .join("\n")];

  _fs.default.writeFileSync(file, data.join("\n"), 'utf-8');

  if (diffRoles) {
    _fs.default.unlinkSync(`${file}.roles`);
  }

  _fs.default.unlinkSync(`${file}.schema`);
};

const apgdiffToFile = (file1, file2, destFile) => {
  let cmd = _env.JAVA_CMD;
  let params = ['-jar', _env.APGDIFF_JAR_PATH, '--add-transaction', file1, file2];
  let options = {};

  if (_env.USE_DOCKER_IMAGE && [_env.SQITCH_CMD, _env.PG_DUMP_CMD, _env.PG_DUMPALL_CMD, _env.JAVA_CMD].indexOf(cmd) !== -1) {
    //alter the command to run in docker
    let w = options && options.cwd ? options.cwd.replace(_env.APP_DIR, _env.DOCKER_APP_DIR) : _env.DOCKER_APP_DIR;
    params = ['run', '--rm', '-w', w, '-v', `${_env.APP_DIR}:${_env.DOCKER_APP_DIR}`, _env.DOCKER_IMAGE, cmd].concat(params.map(p => p.replace(_env.APP_DIR, _env.DOCKER_APP_DIR)));
    cmd = 'docker';
  }

  console.log(`Diffing ${file1} and ${file2}`);
  console.log(`Writing the result to ${destFile}`);

  let p = _child_process.default.spawnSync(cmd, params, options);

  if (p.stdout.toString()) _fs.default.writeFileSync(destFile, p.stdout.toString());
  if (p.stderr.toString()) console.log(p.stderr.toString());
};

const getPgVersion = dbUri => {
  const env = Object.create(process.env),
        result = (0, _common.runCmd)(_env.PSQL_CMD, ['--quiet', '--tuples-only', '-c', 'show server_version', dbUri], {
    env: env
  }, true, false).stdout.toString().trim();
  let pgVersion = '11.3';

  if (result.indexOf(' ') !== -1) {
    pgVersion = result.substr(0, result.indexOf(' '));
  }

  return pgVersion;
};

_commander.default.command('init').option("--debug", "Verbose output and leaves the temporary files (used to create the migration) in place").option("--db-docker-image <image>", "DOcker image used for temp postgres").description('Setup sqitch config and create the first migration').action(options => {
  (0, _common.checkIsAppDir)();
  initMigrations(options.debug, options.dbDockerImage);
});

_commander.default.command('add <name>').option("-n, --note <note>", "Add sqitch migration note").option("-d, --no-diff", "Add empty sqitch migration (no diff)").option("--no-roles", "Do not include ROLE create/drop statements").option("--dry-run", "Don not create migrations files, only output the diff").option("--debug", "Verbose output and leaves the temporary files (used to create the migration) in place").option("--db-uri <uri>", "Diff against a database schema (By default we diff src/ and migrations/deploy/ directories)").option("--db-docker-image <image>", "Docker image used for temp postgres").description('Adds a new sqitch migration').action((name, options) => {
  (0, _common.checkIsAppDir)();
  addMigration(name, options.note, options.diff, options.dryRun, options.debug, options.dbUri, options.dbDockerImage, options.roles);
});

_commander.default.command('sqitch').description('run embeded sqitch').allowUnknownOption().action(() => {
  const sqitchParamas = process.argv.slice(process.argv.indexOf('sqitch') + 1);
  (0, _common.runCmd)(_env.SQITCH_CMD, sqitchParamas, {
    cwd: process.cwd()
  });
}).on('--help', function () {
  (0, _common.runCmd)(_env.SQITCH_CMD, ['--help'], {
    cwd: process.cwd()
  });
});

_commander.default.command('deploy <url>').description('Deploy sqitch migrations to a production database, url must have the `db:pg://${user}:${pass}@${host}:${port}/${db}` format').action(url => {
  (0, _common.checkIsAppDir)();
  (0, _common.checkMigrationsInitiated)();
  (0, _common.sqitchDeploy)(url);
});

_commander.default.parse(process.argv);